/*
 * Copyright (C) 2025 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

#include "usb_libusb.h"

#ifdef ANDROID_TOOLS_USE_BUNDLED_LIBUSB
#include <libusb/libusb.h>
#else
#include <libusb-1.0/libusb.h>
#endif

namespace libusb {

void usb_init();
}

// Only visible to allow inhouse hotplug to inject events.
LIBUSB_CALL int hotplug_callback(libusb_context*, libusb_device* device, libusb_hotplug_event event,
                                 void*);
